@@ -73,6 +73,13 @@ static CRITICAL_SECTION_DEBUG critsect_debug =
 };
 static CRITICAL_SECTION SHELL32_SicCS = { &critsect_debug, -1, 0, 0, 0, 0 };
 
+
+static const WCHAR WindowMetrics[] = {'C','o','n','t','r','o','l',' ','P','a','n','e','l','\\','D','e','s','k','t','o','p','\\',
+                                      'W','i','n','d','o','w','M','e','t','r','i','c','s',0};
+static const WCHAR ShellIconSize[] = {'S','h','e','l','l',' ','I','c','o','n',' ','S','i','z','e',0};
+
+#define SIC_COMPARE_LISTINDEX 1
+
 /*****************************************************************************
  * SIC_CompareEntries
  *
@@ -89,6 +96,10 @@ static INT CALLBACK SIC_CompareEntries( LPVOID p1, LPVOID p2, LPARAM lparam)
 	 * loaded from, their resource index and the fact if they have a shortcut
 	 * icon overlay or not. 
 	 */
+
+        if (lparam & SIC_COMPARE_LISTINDEX)
+            return e1->dwListIndex != e2->dwListIndex;
+
 	if (e1->dwSourceIndex != e2->dwSourceIndex || /* first the faster one */
 	    (e1->dwFlags & GIL_FORSHORTCUT) != (e2->dwFlags & GIL_FORSHORTCUT)) 
 	  return 1;
@@ -99,6 +110,44 @@ static INT CALLBACK SIC_CompareEntries( LPVOID p1, LPVOID p2, LPARAM lparam)
 	return 0;
 }
 
+/**************************************************************************************
+ *                      SIC_get_location
+ *
+ * Returns the source file and resource index of an icon with the given imagelist index
+ */
+HRESULT SIC_get_location( int list_idx, WCHAR *file, DWORD *size, int *res_idx )
+{
+    SIC_ENTRY seek, *found;
+    DWORD needed;
+    HRESULT hr = E_INVALIDARG;
+    int dpa_idx;
+
+    seek.dwListIndex = list_idx;
+
+    EnterCriticalSection( &SHELL32_SicCS );
+
+    dpa_idx = DPA_Search( sic_hdpa, &seek, 0, SIC_CompareEntries, SIC_COMPARE_LISTINDEX, 0 );
+    if (dpa_idx != -1)
+    {
+        found = (SIC_ENTRY *)DPA_GetPtr( sic_hdpa, dpa_idx );
+        needed = (strlenW( found->sSourceFile ) + 1) * sizeof(WCHAR);
+        if (needed <= *size)
+        {
+            memcpy( file, found->sSourceFile, needed );
+            *res_idx = found->dwSourceIndex;
+            hr = S_OK;
+        }
+        else
+        {
+            *size = needed;
+            hr = E_NOT_SUFFICIENT_BUFFER;
+        }
+    }
+    LeaveCriticalSection( &SHELL32_SicCS );
+
+    return hr;
+}
+
 /* declare SIC_LoadOverlayIcon() */
 static int SIC_LoadOverlayIcon(int icon_idx);
 
@@ -361,6 +410,24 @@ static INT SIC_LoadIcon (LPCWSTR sSourceFile, INT dwSourceIndex, DWORD dwFlags)
         return ret;
 }
 
+static int get_shell_icon_size(void)
+{
+    WCHAR buf[32];
+    DWORD value = 32, size = sizeof(buf), type;
+    HKEY key;
+
+    if (!RegOpenKeyW( HKEY_CURRENT_USER, WindowMetrics, &key ))
+    {
+        if (!RegQueryValueExW( key, ShellIconSize, NULL, &type, (BYTE *)buf, &size ) && type == REG_SZ)
+        {
+            if (size == sizeof(buf)) buf[size / sizeof(WCHAR) - 1] = 0;
+            value = atoiW( buf );
+        }
+        RegCloseKey( key );
+    }
+    return value;
+}
+
 /*****************************************************************************
  * SIC_Initialize			[internal]
  */
@@ -370,12 +437,21 @@ static BOOL WINAPI SIC_Initialize( INIT_ONCE *once, void *param, void **context
 	int		cx_small, cy_small;
 	int		cx_large, cy_large;
 
-	cx_small = GetSystemMetrics(SM_CXSMICON);
-	cy_small = GetSystemMetrics(SM_CYSMICON);
-	cx_large = GetSystemMetrics(SM_CXICON);
-	cy_large = GetSystemMetrics(SM_CYICON);
+        if (!IsProcessDPIAware())
+        {
+            cx_large = cy_large = get_shell_icon_size();
+            cx_small = GetSystemMetrics( SM_CXSMICON );
+            cy_small = GetSystemMetrics( SM_CYSMICON );
+        }
+        else
+        {
+            cx_large = GetSystemMetrics( SM_CXICON );
+            cy_large = GetSystemMetrics( SM_CYICON );
+            cx_small = cx_large / 2;
+            cy_small = cy_large / 2;
+        }
 
-	TRACE("\n");
+        TRACE("large %dx%d small %dx%d\n", cx_large, cy_large, cx_small, cx_small);
 
 	sic_hdpa = DPA_Create(16);
 
