@@ -44,8 +44,12 @@ enum wined3d_gl_extension
     APPLE_YCBCR_422,
     /* ARB */
     ARB_BLEND_FUNC_EXTENDED,
+    ARB_CLEAR_BUFFER_OBJECT,
+    ARB_CLEAR_TEXTURE,
     ARB_CLIP_CONTROL,
     ARB_COLOR_BUFFER_FLOAT,
+    ARB_COMPUTE_SHADER,
+    ARB_CONSERVATIVE_DEPTH,
     ARB_COPY_BUFFER,
     ARB_DEBUG_OUTPUT,
     ARB_DEPTH_BUFFER_FLOAT,
@@ -58,11 +62,13 @@ enum wined3d_gl_extension
     ARB_ES3_COMPATIBILITY,
     ARB_EXPLICIT_ATTRIB_LOCATION,
     ARB_FRAGMENT_COORD_CONVENTIONS,
+    ARB_FRAGMENT_LAYER_VIEWPORT,
     ARB_FRAGMENT_PROGRAM,
     ARB_FRAGMENT_SHADER,
     ARB_FRAMEBUFFER_OBJECT,
     ARB_FRAMEBUFFER_SRGB,
     ARB_GEOMETRY_SHADER4,
+    ARB_GPU_SHADER5,
     ARB_HALF_FLOAT_PIXEL,
     ARB_HALF_FLOAT_VERTEX,
     ARB_INSTANCED_ARRAYS,
@@ -73,27 +79,38 @@ enum wined3d_gl_extension
     ARB_MULTISAMPLE,
     ARB_MULTITEXTURE,
     ARB_OCCLUSION_QUERY,
+    ARB_PIPELINE_STATISTICS_QUERY,
     ARB_PIXEL_BUFFER_OBJECT,
     ARB_POINT_PARAMETERS,
     ARB_POINT_SPRITE,
     ARB_PROVOKING_VERTEX,
     ARB_SAMPLER_OBJECTS,
+    ARB_SEAMLESS_CUBE_MAP,
+    ARB_SHADER_ATOMIC_COUNTERS,
     ARB_SHADER_BIT_ENCODING,
     ARB_SHADER_IMAGE_LOAD_STORE,
     ARB_SHADER_IMAGE_SIZE,
+    ARB_SHADER_STORAGE_BUFFER_OBJECT,
     ARB_SHADER_TEXTURE_LOD,
     ARB_SHADING_LANGUAGE_100,
+    ARB_SHADING_LANGUAGE_420PACK,
+    ARB_SHADING_LANGUAGE_PACKING,
     ARB_SHADOW,
     ARB_STENCIL_TEXTURING,
     ARB_SYNC,
+    ARB_TESSELLATION_SHADER,
     ARB_TEXTURE_BORDER_CLAMP,
+    ARB_TEXTURE_BUFFER_OBJECT,
+    ARB_TEXTURE_BUFFER_RANGE,
     ARB_TEXTURE_COMPRESSION,
     ARB_TEXTURE_COMPRESSION_BPTC,
     ARB_TEXTURE_COMPRESSION_RGTC,
     ARB_TEXTURE_CUBE_MAP,
+    ARB_TEXTURE_CUBE_MAP_ARRAY,
     ARB_TEXTURE_ENV_COMBINE,
     ARB_TEXTURE_ENV_DOT3,
     ARB_TEXTURE_FLOAT,
+    ARB_TEXTURE_GATHER,
     ARB_TEXTURE_MIRRORED_REPEAT,
     ARB_TEXTURE_MIRROR_CLAMP_TO_EDGE,
     ARB_TEXTURE_NON_POWER_OF_TWO,
@@ -105,6 +122,8 @@ enum wined3d_gl_extension
     ARB_TEXTURE_SWIZZLE,
     ARB_TEXTURE_VIEW,
     ARB_TIMER_QUERY,
+    ARB_TRANSFORM_FEEDBACK2,
+    ARB_TRANSFORM_FEEDBACK3,
     ARB_UNIFORM_BUFFER_OBJECT,
     ARB_VERTEX_ARRAY_BGRA,
     ARB_VERTEX_BLEND,
@@ -150,6 +169,7 @@ enum wined3d_gl_extension
     EXT_TEXTURE_INTEGER,
     EXT_TEXTURE_LOD_BIAS,
     EXT_TEXTURE_MIRROR_CLAMP,
+    EXT_TEXTURE_SHARED_EXPONENT,
     EXT_TEXTURE_SNORM,
     EXT_TEXTURE_SRGB,
     EXT_TEXTURE_SRGB_DECODE,
@@ -183,11 +203,12 @@ enum wined3d_gl_extension
     WGL_WINE_QUERY_RENDERER,
     /* Internally used */
     WINED3D_GL_BLEND_EQUATION,
-    WINED3D_GL_NORMALIZED_TEXRECT,
     WINED3D_GL_LEGACY_CONTEXT,
+    WINED3D_GL_NORMALIZED_TEXRECT,
+    WINED3D_GL_PRIMITIVE_QUERY,
     WINED3D_GL_VERSION_2_0,
     WINED3D_GL_VERSION_3_2,
-    WINED3D_GL_VERSION_4_3,
+    WINED3D_GLSL_130,
 
     WINED3D_GL_EXT_COUNT,
 };
